import os
import logging
import yaml
from openai import AsyncOpenAI

from logdetective.utils import load_prompts, load_skip_snippet_patterns
from logdetective.server.models import Config, InferenceConfig
import logdetective


def load_server_config(path: str | None) -> Config:
    """Load configuration file for logdetective server.
    If no path was provided, or if the file doesn't exist, return defaults.
    """
    if path is not None:
        try:
            with open(path, "r") as config_file:
                return Config(yaml.safe_load(config_file))
        except FileNotFoundError:
            # This is not an error, we will fall back to default
            print("Unable to find server config file, using default then.")
    return Config()


def get_log(config: Config):
    """
    Initialize a logger for this server
    """
    log = logging.getLogger(config.log.name)
    if getattr(log, "initialized", False):
        return log

    log.setLevel("DEBUG")

    # Drop the default handler, we will create it ourselves
    log.handlers = []

    # STDOUT
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter(config.log.format))
    stream_handler.setLevel(config.log.level_stream)
    log.addHandler(stream_handler)

    # Log to file
    if config.log.path:
        file_handler = logging.FileHandler(config.log.path)
        file_handler.setFormatter(logging.Formatter(config.log.format))
        file_handler.setLevel(config.log.level_file)
        log.addHandler(file_handler)

    log.initialized = True
    return log


def get_openai_api_client(inference_config: InferenceConfig):
    """Set up AsyncOpenAI client with default configuration."""
    return AsyncOpenAI(
        api_key=inference_config.api_token, base_url=inference_config.url,
        timeout=inference_config.llm_api_timeout
    )


SERVER_CONFIG_PATH = os.environ.get("LOGDETECTIVE_SERVER_CONF", None)
SERVER_PROMPT_PATH = os.environ.get("LOGDETECTIVE_PROMPTS", None)
# The default location for skip patterns is in the same directory
# as logdetective __init__.py file.
SERVER_SKIP_PATTERNS_PATH = os.environ.get(
    "LOGDETECIVE_SKIP_PATTERNS",
    f"{os.path.dirname(logdetective.__file__)}/skip_snippets.yml",
)

SERVER_CONFIG = load_server_config(SERVER_CONFIG_PATH)
PROMPT_CONFIG = load_prompts(SERVER_PROMPT_PATH)
SKIP_SNIPPETS_CONFIG = load_skip_snippet_patterns(SERVER_SKIP_PATTERNS_PATH)

LOG = get_log(SERVER_CONFIG)

CLIENT = get_openai_api_client(SERVER_CONFIG.inference)



import asyncio

from typing import List, Callable
from collections import Counter

import gitlab

from logdetective.server.models import TimePeriod
from logdetective.server.database.models import (
    Comments,
    Reactions,
    GitlabMergeRequestJobs,
    Forge,
)
from logdetective.server.config import LOG


async def collect_emojis(gitlab_conn: gitlab.Gitlab, period: TimePeriod):
    """
    Collect emoji feedback from logdetective comments saved in database.
    Check only comments created in the last given period of time.
    """
    comments = await Comments.get_since(period.get_period_start_time()) or []
    comments_for_gitlab_connection = [
        comment for comment in comments if comment.forge == gitlab_conn.url
    ]
    await collect_emojis_in_comments(comments_for_gitlab_connection, gitlab_conn)


async def collect_emojis_for_mr(
    project_id: int, mr_iid: int, gitlab_conn: gitlab.Gitlab
):
    """
    Collect emoji feedback from logdetective comments in the specified MR.
    """
    comments = []
    try:
        url = Forge(gitlab_conn.url)
    except ValueError as ex:
        LOG.exception("Attempt to use unrecognized Forge `%s`", gitlab_conn.url)
        raise ex
    mr_jobs = await GitlabMergeRequestJobs.get_by_mr_iid(url, project_id, mr_iid) or []

    comments = [await Comments.get_by_mr_job(mr_job) for mr_job in mr_jobs]
    await collect_emojis_in_comments(comments, gitlab_conn)


async def _handle_gitlab_operation(func: Callable, *args):
    """
    It handles errors for the specified GitLab operation.
    After executing it in a separate thread.
    """
    try:
        return await asyncio.to_thread(func, *args)
    except (gitlab.GitlabError, gitlab.GitlabGetError) as e:
        log_msg = f"Error during GitLab operation {func}{args}: {e}"
        if "Not Found" in str(e):
            LOG.error(log_msg)
        else:
            LOG.exception(log_msg)
    except Exception as e:  # pylint: disable=broad-exception-caught
        LOG.exception(
            "Unexpected error during GitLab operation %s(%s): %s", func, args, e
        )


async def collect_emojis_in_comments(  # pylint: disable=too-many-locals
    comments: List[Comments], gitlab_conn: gitlab.Gitlab
):
    """
    Collect emoji feedback from specified logdetective comments.
    """
    projects = {}
    merge_requests = {}
    for comment in comments:
        mr_job_db = await GitlabMergeRequestJobs.get_by_id(comment.merge_request_job_id)
        if not mr_job_db:
            continue
        if mr_job_db.id not in projects:
            project = await _handle_gitlab_operation(
                gitlab_conn.projects.get, mr_job_db.project_id
            )
            if not project:
                continue
            projects[mr_job_db.id] = project
        else:
            project = projects[mr_job_db.id]
        merge_request_iid = mr_job_db.mr_iid
        if merge_request_iid not in merge_requests:
            merge_request = await _handle_gitlab_operation(
                project.mergerequests.get, merge_request_iid
            )
            if not merge_request:
                continue
            merge_requests[merge_request_iid] = merge_request
        else:
            merge_request = merge_requests[merge_request_iid]

        discussion = await _handle_gitlab_operation(
            merge_request.discussions.get, comment.comment_id
        )
        if not discussion:
            continue

        # Get the ID of the first note
        note_id = discussion.attributes["notes"][0]["id"]
        note = await _handle_gitlab_operation(merge_request.notes.get, note_id)
        if not note:
            continue

        emoji_counts = Counter(emoji.name for emoji in note.awardemojis.list())

        # keep track of not updated reactions
        # because we need to remove them
        old_emojis = [
            reaction.reaction_type
            for reaction in await Reactions.get_all_reactions(
                comment.forge,
                mr_job_db.project_id,
                mr_job_db.mr_iid,
                mr_job_db.job_id,
                comment.comment_id,
            )
        ]
        for key, value in emoji_counts.items():
            await Reactions.create_or_update(
                comment.forge,
                mr_job_db.project_id,
                mr_job_db.mr_iid,
                mr_job_db.job_id,
                comment.comment_id,
                key,
                value,
            )
            if key in old_emojis:
                old_emojis.remove(key)

        # not updated reactions has been removed, drop them
        await Reactions.delete(
            comment.forge,
            mr_job_db.project_id,
            mr_job_db.mr_iid,
            mr_job_db.job_id,
            comment.comment_id,
            old_emojis,
        )



import argparse
import asyncio
import logging
import sys
import os

import aiohttp

from logdetective.constants import DEFAULT_ADVISOR, DEFAULT_TEMPERATURE
from logdetective.utils import (
    process_log,
    initialize_model,
    retrieve_log_content,
    format_snippets,
    compute_certainty,
    load_prompts,
    load_skip_snippet_patterns,
    check_csgrep,
    mine_logs,
)
from logdetective.extractors import DrainExtractor, CSGrepExtractor

LOG = logging.getLogger("logdetective")


def setup_args():
    """Setup argument parser and return arguments."""
    parser = argparse.ArgumentParser("logdetective")
    parser.add_argument(
        "file",
        type=str,
        default="",
        help="The URL or path to the log file to be analyzed.",
    )
    parser.add_argument(
        "-M",
        "--model",
        help="The path or Hugging Face name of the language model for analysis.",
        type=str,
        default=DEFAULT_ADVISOR,
    )
    parser.add_argument(
        "-F",
        "--filename_suffix",
        help="Suffix of the model file name to be retrieved from Hugging Face.\
                            Makes sense only if the model is specified with Hugging Face name.",
        default="Q4_K.gguf",
    )
    parser.add_argument("-n", "--no-stream", action="store_true")
    parser.add_argument(
        "-S",
        "--summarizer",
        type=str,
        default="drain",
        help="DISABLED: LLM summarization option was removed. \
                Argument is kept for backward compatibility only.",
    )
    parser.add_argument(
        "-N",
        "--n_lines",
        type=int,
        default=None,
        help="DISABLED: LLM summarization option was removed. \
                Argument is kept for backward compatibility only.",
    )
    parser.add_argument(
        "-C",
        "--n_clusters",
        type=int,
        default=8,
        help="Number of clusters for Drain to organize log chunks into.\
                            This only makes sense when you are summarizing with Drain",
    )
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("-q", "--quiet", action="store_true")
    parser.add_argument(
        "--prompts",
        type=str,
        default=f"{os.path.dirname(__file__)}/prompts.yml",
        help="Path to prompt configuration file.",
    )
    parser.add_argument(
        "--temperature",
        type=float,
        default=DEFAULT_TEMPERATURE,
        help="Temperature for inference.",
    )
    parser.add_argument(
        "--skip_snippets",
        type=str,
        default=f"{os.path.dirname(__file__)}/skip_snippets.yml",
        help="Path to patterns for skipping snippets.",
    )
    parser.add_argument(
        "--csgrep", action="store_true", help="Use csgrep to process the log."
    )
    return parser.parse_args()


async def run():  # pylint: disable=too-many-statements,too-many-locals,too-many-branches
    """Main execution function."""
    args = setup_args()

    if args.verbose and args.quiet:
        sys.stderr.write("Error: --quiet and --verbose is mutually exclusive.\n")
        sys.exit(2)

    # Emit warning about use of discontinued args
    if args.n_lines or args.summarizer != "drain":
        LOG.warning("LLM based summarization was removed. Drain will be used instead.")

    # Logging facility setup
    log_level = logging.INFO
    if args.verbose >= 1:
        log_level = logging.DEBUG
    if args.quiet:
        log_level = 0

    # Get prompts configuration
    prompts_configuration = load_prompts(args.prompts)

    logging.basicConfig(stream=sys.stdout)
    LOG.setLevel(log_level)

    # Primary model initialization
    try:
        model = initialize_model(
            args.model, filename_suffix=args.filename_suffix, verbose=args.verbose > 2
        )
    except ValueError as e:
        LOG.error(e)
        LOG.error("You likely do not have enough memory to load the AI model")
        sys.exit(3)

    try:
        skip_snippets = load_skip_snippet_patterns(args.skip_snippets)
    except OSError as e:
        LOG.error(e)
        sys.exit(5)

    # Log file summarizer initialization
    extractors = []
    extractors.append(
        DrainExtractor(
            args.verbose > 1,
            max_clusters=args.n_clusters,
            skip_snippets=skip_snippets,
        )
    )

    if args.csgrep:
        if not check_csgrep():
            LOG.error(
                "You have requested use of `csgrep` when it isn't available on your system."
            )
            sys.exit(6)
        extractors.append(
            CSGrepExtractor(args.verbose > 1, skip_snippets=skip_snippets)
        )

    LOG.info("Getting summary")

    async with aiohttp.ClientSession() as http:
        try:
            log = await retrieve_log_content(http, args.file)
        except ValueError as e:
            # file does not exist
            LOG.error(e)
            sys.exit(4)

    log_summary = mine_logs(log=log, extractors=extractors)
    LOG.info("Analyzing the text")

    log_summary = format_snippets(log_summary)
    LOG.info("Log summary: \n %s", log_summary)

    prompt = (
        f"{prompts_configuration.default_system_prompt}\n"
        f"{prompts_configuration.prompt_template}"
    )

    stream = True
    if args.no_stream:
        stream = False
    response = process_log(
        log_summary,
        model,
        stream,
        prompt_template=prompt,
        temperature=args.temperature,
    )
    probs = []
    print("Explanation:")
    # We need to extract top token probability from the response
    # CreateCompletionResponse structure of llama-cpp-python.
    # `compute_certainty` function expects list of dictionaries with form
    # { 'logprob': <float> } as expected from the OpenAI API.

    if args.no_stream:
        print(response["choices"][0]["text"])
        probs = [
            {"logprob": e} for e in response["choices"][0]["logprobs"]["token_logprobs"]
        ]

    else:
        # Stream the output
        for chunk in response:
            if isinstance(chunk["choices"][0]["logprobs"], dict):
                probs.append(
                    {"logprob": chunk["choices"][0]["logprobs"]["token_logprobs"][0]}
                )
            delta = chunk["choices"][0]["text"]
            print(delta, end="", flush=True)
    certainty = compute_certainty(probs)

    print(f"\nResponse certainty: {certainty:.2f}%\n")


def main():
    """Evaluate logdetective program and wait for it to finish"""
    asyncio.run(run())


if __name__ == "__main__":
    main()



import logging
import os
import subprocess as sp
from typing import Iterator, List, Dict, Tuple, Generator
from urllib.parse import urlparse

import aiohttp
import numpy as np
import yaml

from llama_cpp import Llama, CreateCompletionResponse, CreateCompletionStreamResponse
from logdetective.constants import SNIPPET_DELIMITER
from logdetective.models import PromptConfig, SkipSnippets
from logdetective.remote_log import RemoteLog

LOG = logging.getLogger("logdetective")


def new_message(text: str) -> bool:
    """Set of heuristics for determining whether or not
    does the current chunk of log text continue on next line.

    Following rules are checked, in order:
    * is the first character is whitespace
    * is the first character backslash '|'
    """
    conditionals = [
        lambda string: string[0].isspace(),
        lambda string: string[0] == "|",
    ]

    for c in conditionals:
        y = c(text)
        if y:
            return False

    return True


def get_chunks(
    text: str, max_chunk_len: int = 2000
) -> Generator[Tuple[int, str], None, None]:
    """Split log into chunks according to heuristic
    based on whitespace and backslash presence.
    """
    lines = text.splitlines()

    # Chunk we will be yielding
    chunk = ""
    # Number of line where the message started
    original_line = 0
    for i, line in enumerate(lines):
        if len(line) == 0:
            continue
        if new_message(line):
            # Yield chunk if we have it
            if len(chunk) > 0:
                yield (original_line, chunk)
            original_line = i
            chunk = line
        else:
            chunk += "\n" + line
        if len(chunk) > max_chunk_len:
            # If the chunk is too long, keep splitting into smaller chunks
            # until we reach manageable size
            while len(chunk) > max_chunk_len:
                remainder = chunk[max_chunk_len:]
                chunk = chunk[:max_chunk_len]
                yield (original_line, chunk)
                chunk = remainder

    # if we still have some text left over
    yield (original_line, chunk)


def initialize_model(
    model_pth: str, filename_suffix: str = ".gguf", verbose: bool = False
) -> Llama:
    """Initialize Llama class for inference.
    Args:
        model_pth (str): path to gguf model file or Hugging Face name
        filename_suffix (str): suffix of the model file name to be pulled from Hugging Face
        verbose (bool): level of verbosity for llamacpp
    """

    LOG.info("Loading model from %s", model_pth)

    if os.path.isfile(model_pth):
        model = Llama(
            model_path=model_pth,
            n_ctx=0,  # Maximum context for the model
            verbose=verbose,
            logits_all=True,
        )
    else:
        model = Llama.from_pretrained(
            model_pth,
            f"*{filename_suffix}",
            n_ctx=0,  # Maximum context for the model
            verbose=verbose,
            logits_all=True,
        )

    return model


def compute_certainty(probs: List[Dict]) -> float:
    """Compute certainty of repsponse based on average logit probability.
    Log probability is log(p), isn't really readable for most people, especially in compound.
    In this case it's just a matter of applying inverse operation exp.
    Of course that leaves you with a value in range <0, 1> so it needs to be multiplied by 100.
    Simply put, this is the most straightforward way to get the numbers out.

    This function is used in the server codebase.
    """

    top_logprobs = [np.exp(e["logprob"]) * 100 for e in probs]

    certainty = np.median(top_logprobs, axis=0)
    if np.isnan(certainty):
        raise ValueError("NaN certainty of answer")
    return certainty


def process_log(
    log: str, model: Llama, stream: bool, prompt_template: str, temperature: float
) -> CreateCompletionResponse | Iterator[CreateCompletionStreamResponse]:
    """Processes a given log using the provided language model and returns its summary.

    Args:
        log (str): The input log to be processed.
        model (Llama): The language model used for processing the log.
        stream (bool): Return output as Iterator.
        prompt_template (str): Which prompt template to use.
        temperature (float): Temperature parameter for model runtime.
    Returns:
        str: The summary of the given log generated by the language model.
    """
    response = model(
        prompt=prompt_template.format(log),
        stream=stream,
        max_tokens=0,
        logprobs=1,
        temperature=temperature,
    )

    return response


async def retrieve_log_content(http: aiohttp.ClientSession, log_path: str) -> str:
    """Get content of the file on the log_path path.
    Path is assumed to be valid URL if it has a scheme.
    Otherwise it attempts to pull it from local filesystem."""
    parsed_url = urlparse(log_path)
    log = ""

    if not parsed_url.scheme:
        if not os.path.exists(log_path):
            raise ValueError(f"Local log {log_path} doesn't exist!")

        with open(log_path, "rt") as f:
            log = f.read()

    else:
        remote_log = RemoteLog(log_path, http)
        log = await remote_log.get_url_content()

    return log


def format_snippets(snippets: list[str] | list[Tuple[int, str]]) -> str:
    """Format snippets, giving them separator, id and finally
    concatenating them. If snippets have line number attached,
    include that in prompt.

    Line number must be first element in the tuple. Mixed format of snippets
    is permitted, but may have impact on inference.
    """
    summary = ""
    for i, s in enumerate(snippets):
        if isinstance(s, tuple):
            summary += f"""
            Snippet No. {i} at line #{s[0]}:

            {s[1]}
            {SNIPPET_DELIMITER}
            """
        else:
            summary += f"""
            Snippet No. {i}:

            {s}
            {SNIPPET_DELIMITER}
            """
    return summary


def load_prompts(path: str | None) -> PromptConfig:
    """Load prompts from given yaml file if there is one.
    Alternatively use defaults."""
    if path:
        try:
            with open(path, "r") as file:
                return PromptConfig(yaml.safe_load(file))
        except FileNotFoundError:
            print("Prompt configuration file not found, reverting to defaults.")
    return PromptConfig()


def prompt_to_messages(
    user_message: str,
    system_prompt: str | None = None,
    system_role: str = "developer",
    user_role: str = "user",
) -> List[Dict[str, str]]:
    """Turn prompt into list of message dictionaries.
    If `system_role` and `user_role` are the same, only a single message is created,
    as concatenation of `user_message` and `system_prompt`. This is useful for models which
    do not have separate system role, such as mistral.
    """

    if system_role == user_role:
        messages = [
            {"role": system_role, "content": f"{system_prompt}\n{user_message}"}
        ]
    else:
        messages = [
            {"role": system_role, "content": system_prompt},
            {
                "role": user_role,
                "content": user_message,
            },
        ]

    return messages


def filter_snippet_patterns(snippet: str, skip_snippets: SkipSnippets) -> bool:
    """Try to match snippet agains provided patterns to determine if we should
    filter it out or not."""
    for key, pattern in skip_snippets.snippet_patterns.items():
        if pattern.match(snippet):
            LOG.debug("Snippet `%s` has matched agains skip pattern %s", snippet, key)
            return True

    return False


def load_skip_snippet_patterns(path: str | None) -> SkipSnippets:
    """Load dictionary of snippet patterns we want to skip."""
    if path:
        try:
            with open(path, "r") as file:
                return SkipSnippets(yaml.safe_load(file))
        except OSError as e:
            LOG.error("Couldn't open file with snippet skip patterns `%s`", path)
            raise e

    return SkipSnippets({})


def check_csgrep() -> bool:
    """Verifies presence of csgrep in path"""
    try:
        result = sp.run(
            ["csgrep", "--version"],
            text=True,
            check=True,
            shell=False,
            capture_output=True,
            timeout=1.0,
        )
    except (FileNotFoundError, sp.TimeoutExpired, sp.CalledProcessError) as ex:
        LOG.error("Required binary `csgrep` was not found in path: %s", ex)
        return False
    if result.returncode == 0:
        return True
    LOG.error("Issue was encountered while calling `csgrep`: `%s`", result.stderr)

    return False


def mine_logs(log: str, extractors: list) -> List[Tuple[int, str]]:
    """Extract snippets from log text using extractors provided.
    Each extractor is applied in turn on original log.
    Depending on characteristics of extractors used, there may be
    an overlap in snippets extracted."""

    log_summary = []

    LOG.info("Getting summary")

    for extractor in extractors:
        log_summary.extend(extractor(log))

    ratio = len("\n".join([text for _, text in log_summary])) / len(log)
    LOG.debug("Log summary: \n %s", log_summary)
    LOG.info("Snippets: %s Compression ratio: %s", len(log_summary), ratio)

    return log_summary



import datetime
from typing import Optional, Union, Dict

import numpy
from numpy.typing import ArrayLike
from matplotlib import dates, colormaps, axes, pyplot, figure

from logdetective.server.models import TimePeriod
from logdetective.server.database.models import (
    AnalyzeRequestMetrics,
    EndpointType,
    Reactions,
)


class Definition:
    """Define plot details, given a time period."""

    def __init__(self, time_period: TimePeriod):
        self.time_period = time_period
        self.days_diff = time_period.get_time_period().days
        if self.time_period.hours:
            self._freq = "H"
            self._time_format = "%Y-%m-%d %H"
            self._locator = dates.HourLocator(interval=2)
            self._time_unit = "hour"
            self._time_delta = datetime.timedelta(hours=1)
        elif self.time_period.days:
            self._freq = "D"
            self._time_format = "%Y-%m-%d"
            self._locator = dates.DayLocator(interval=1)
            self._time_unit = "day"
            self._time_delta = datetime.timedelta(days=1)
        elif self.time_period.weeks:
            self._freq = "W"
            self._time_format = "%Y-%m-%d"
            self._locator = dates.WeekdayLocator(interval=1)
            self._time_unit = "week"
            self._time_delta = datetime.timedelta(weeks=1)

    @property
    def freq(self):
        # pylint: disable=missing-function-docstring
        return self._freq

    @property
    def time_format(self):
        # pylint: disable=missing-function-docstring
        return self._time_format

    @property
    def locator(self):
        # pylint: disable=missing-function-docstring
        return self._locator

    @property
    def time_unit(self):
        # pylint: disable=missing-function-docstring
        return self._time_unit

    @property
    def time_delta(self):
        # pylint: disable=missing-function-docstring
        return self._time_delta


def create_time_series_arrays(
    values_dict: dict[datetime.datetime, int],
    plot_def: Definition,
    start_time: datetime.datetime,
    end_time: datetime.datetime,
    value_type: Optional[Union[int, float]] = int,
) -> tuple[numpy.ndarray, numpy.ndarray]:
    """Create time series arrays from a dictionary of values.

    This function generates two aligned numpy arrays:
    1. An array of timestamps from start_time to end_time
    2. A corresponding array of valuesfor each timestamp

    The timestamps are truncated to the precision specified by time_format.
    If a timestamp in values_dict matches a generated timestamp, its values is used;
    otherwise, the value defaults to zero.

    Args:
        values_dict: Dictionary mapping timestamps to their respective values
        start_time: The starting timestamp of the time series
        end_time: The ending timestamp of the time series
        time_delta: The time interval between consecutive timestamps
        time_format: String format for datetime truncation (e.g., '%Y-%m-%d %H:%M')

    Returns:
        A tuple containing:
            - numpy.ndarray: Array of timestamps
            - numpy.ndarray: Array of corresponding values
    """
    num_intervals = int((end_time - start_time) / plot_def.time_delta) + 1

    timestamps = numpy.array(
        [
            datetime.datetime.strptime(
                (start_time + i * plot_def.time_delta).strftime(
                    format=plot_def.time_format
                ),
                plot_def.time_format,
            )
            for i in range(num_intervals)
        ]
    )
    values = numpy.zeros(num_intervals, dtype=value_type)

    timestamp_to_index = {timestamp: i for i, timestamp in enumerate(timestamps)}

    for timestamp, count in values_dict.items():
        if timestamp in timestamp_to_index:
            values[timestamp_to_index[timestamp]] = count

    return timestamps, values


def _add_bar_chart(
    ax: axes.Axes,
    plot_def: Definition,
    timestamps: ArrayLike,
    values: ArrayLike,
    label: str,
) -> None:
    """Add a blue bar chart"""
    bar_width = (
        0.8 * plot_def.time_delta.total_seconds() / 86400
    )  # Convert to days for matplotlib
    ax.bar(
        timestamps,
        values,
        width=bar_width,
        alpha=0.7,
        color="skyblue",
        label=label,
    )
    ax.set_xlabel("Time")
    ax.set_ylabel(label, color="blue")
    ax.tick_params(axis="y", labelcolor="blue")

    ax.xaxis.set_major_formatter(dates.DateFormatter(plot_def.time_format))
    ax.xaxis.set_major_locator(plot_def.locator)

    pyplot.xticks(rotation=45)

    ax.grid(True, alpha=0.3)


def _add_line_chart(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
    ax: axes.Axes,
    timestamps: ArrayLike,
    values: ArrayLike,
    label: str,
    color: str = "red",
    set_label: bool = True,
):
    """Add a red line chart"""
    ax.plot(timestamps, values, color=color, linestyle="-", linewidth=2, label=label)
    if set_label:
        ax.set_ylabel(label, color=color)
        ax.tick_params(axis="y", labelcolor=color)


async def requests_per_time(
    period_of_time: TimePeriod,
    endpoint: EndpointType = EndpointType.ANALYZE,
    end_time: Optional[datetime.datetime] = None,
) -> figure.Figure:
    """
    Generate a visualization of request counts over a specified time period.

    This function creates a dual-axis plot showing:
    1. A bar chart of request counts per time interval
    2. A line chart showing the cumulative request count

    The time intervals are determined by the provided TimePeriod object, which defines
    the granularity and formatting of the time axis.

    Args:
        period_of_time: A TimePeriod object that defines the time period and interval
                        for the analysis (e.g., hourly, daily, weekly)
        endpoint: One of the API endpoints
        end_time: The end time for the analysis period. If None, defaults to the current
                  UTC time

    Returns:
        A matplotlib Figure object containing the generated visualization
    """
    end_time = end_time or datetime.datetime.now(datetime.timezone.utc)
    start_time = period_of_time.get_period_start_time(end_time)
    plot_def = Definition(period_of_time)
    requests_counts = await AnalyzeRequestMetrics.get_requests_in_period(
        start_time, end_time, plot_def.time_format, endpoint
    )
    timestamps, counts = create_time_series_arrays(
        requests_counts, plot_def, start_time, end_time
    )

    fig, ax1 = pyplot.subplots(figsize=(12, 6))
    _add_bar_chart(ax1, plot_def, timestamps, counts, "Requests")

    ax2 = ax1.twinx()
    _add_line_chart(ax2, timestamps, numpy.cumsum(counts), "Cumulative Requests")

    pyplot.title(
        f"Requests received for API {endpoint} ({start_time.strftime(plot_def.time_format)} "
        f"to {end_time.strftime(plot_def.time_format)})"
    )

    lines1, labels1 = ax1.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    ax1.legend(lines1 + lines2, labels1 + labels2, loc="center")

    pyplot.tight_layout()

    return fig


async def average_time_per_responses(  # pylint: disable=too-many-locals
    period_of_time: TimePeriod,
    endpoint: EndpointType = EndpointType.ANALYZE,
    end_time: Optional[datetime.datetime] = None,
) -> figure.Figure:
    """
    Generate a visualization of average response time and length over a specified time period.

    This function creates a dual-axis plot showing:
    1. A bar chart of average response time per time interval
    1. A line chart of average response length per time interval

    The time intervals are determined by the provided TimePeriod object, which defines
    the granularity and formatting of the time axis.

    Args:
        period_of_time: A TimePeriod object that defines the time period and interval
                        for the analysis (e.g., hourly, daily, weekly)
        endpoint: One of the API endpoints
        end_time: The end time for the analysis period. If None, defaults to the current
                  UTC time

    Returns:
        A matplotlib Figure object containing the generated visualization
    """
    end_time = end_time or datetime.datetime.now(datetime.timezone.utc)
    start_time = period_of_time.get_period_start_time(end_time)
    plot_def = Definition(period_of_time)
    responses_average_time = (
        await AnalyzeRequestMetrics.get_responses_average_time_in_period(
            start_time, end_time, plot_def.time_format, endpoint
        )
    )
    timestamps, average_time = create_time_series_arrays(
        responses_average_time,
        plot_def,
        start_time,
        end_time,
        float,
    )

    fig, ax1 = pyplot.subplots(figsize=(12, 6))
    _add_bar_chart(
        ax1, plot_def, timestamps, average_time, "average response time (seconds)"
    )

    responses_average_length = (
        await AnalyzeRequestMetrics.get_responses_average_length_in_period(
            start_time, end_time, plot_def.time_format, endpoint
        )
    )
    timestamps, average_length = create_time_series_arrays(
        responses_average_length,
        plot_def,
        start_time,
        end_time,
        float,
    )

    ax2 = ax1.twinx()
    _add_line_chart(ax2, timestamps, average_length, "average response length (chars)")

    pyplot.title(
        f"average response time for API {endpoint} ({start_time.strftime(plot_def.time_format)} "
        f"to {end_time.strftime(plot_def.time_format)})"
    )

    lines1, labels1 = ax1.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    ax1.legend(lines1 + lines2, labels1 + labels2, loc="center")

    pyplot.tight_layout()

    return fig


async def _collect_emoji_data(
    start_time: datetime.datetime, plot_def: Definition
) -> Dict[str, Dict[datetime.datetime, int]]:
    """Collect and organize emoji feedback data

    Counts all emojis given to logdetective comments created since start_time.
    Collect counts in time accordingly to the plot definition.
    """
    reactions = await Reactions.get_since(start_time)
    reactions_values_dict: Dict[str, Dict] = {}
    for comment_created_at, reaction in reactions:
        comment_created_at_formatted = comment_created_at.strptime(
            comment_created_at.strftime(plot_def.time_format), plot_def.time_format
        )
        if reaction.reaction_type in reactions_values_dict:
            reaction_values_dict = reactions_values_dict[reaction.reaction_type]
            if comment_created_at_formatted in reaction_values_dict:
                reaction_values_dict[comment_created_at_formatted] += reaction.count
            else:
                reaction_values_dict[comment_created_at_formatted] = reaction.count
        else:
            reaction_values_dict = {comment_created_at_formatted: reaction.count}
            reactions_values_dict.update({reaction.reaction_type: reaction_values_dict})

    return reactions_values_dict


def _plot_emoji_data(  # pylint: disable=too-many-locals
    ax: axes.Axes,
    reactions_values_dict: Dict[str, Dict[datetime.datetime, int]],
    plot_def: Definition,
    start_time: datetime.datetime,
    end_time: datetime.datetime,
):
    """Plot each emoji's data on its own axis."""
    emoji_lines = {}
    emoji_labels = {}

    # Find global min and max y values to set consistent scale
    all_counts = []
    for emoji, dict_counts in reactions_values_dict.items():
        timestamps, counts = create_time_series_arrays(
            dict_counts, plot_def, start_time, end_time
        )
        all_counts.extend(counts)

    colors = [
        colormaps["viridis"](i)
        for i in numpy.linspace(0, 1, len(reactions_values_dict))
    ]

    first_emoji = True
    for i, (emoji, dict_counts) in enumerate(reactions_values_dict.items()):
        timestamps, counts = create_time_series_arrays(
            dict_counts, plot_def, start_time, end_time
        )

        if first_emoji:
            current_ax = ax
            first_emoji = False
        else:
            current_ax = ax.twinx()
            current_ax.spines["right"].set_position(("outward", 60 * (i - 1)))

        _add_line_chart(current_ax, timestamps, counts, f"{emoji}", colors[i], False)
        emoji_lines[emoji], emoji_labels[emoji] = current_ax.get_legend_handles_labels()

        # Set the same y-limits for all axes
        current_ax.set_ylim(0, max(all_counts) * 1.1)

        # Only show y-ticks on the first axis to avoid clutter
        if 0 < i < len(reactions_values_dict):
            current_ax.set_yticks([])

    return emoji_lines, emoji_labels


async def emojis_per_time(
    period_of_time: TimePeriod,
    end_time: Optional[datetime.datetime] = None,
) -> figure.Figure:
    """
    Generate a visualization of overall emoji feedback
    over a specified time period.

    This function creates a multiple-axis plot showing
    a line chart for every found emoji

    The time intervals are determined by the provided TimePeriod object, which defines
    the granularity and formatting of the time axis.

    Args:
        period_of_time: A TimePeriod object that defines the time period and interval
                        for the analysis (e.g., hourly, daily, weekly)
        end_time: The end time for the analysis period. If None, defaults to the current
                  UTC time

    Returns:
        A matplotlib Figure object containing the generated visualization
    """
    plot_def = Definition(period_of_time)
    end_time = end_time or datetime.datetime.now(datetime.timezone.utc)
    start_time = period_of_time.get_period_start_time(end_time)
    reactions_values_dict = await _collect_emoji_data(start_time, plot_def)

    fig, ax = pyplot.subplots(figsize=(12, 6))

    emoji_lines, emoji_labels = _plot_emoji_data(
        ax, reactions_values_dict, plot_def, start_time, end_time
    )

    pyplot.title(
        f"Emoji feedback ({start_time.strftime(plot_def.time_format)} "
        f"to {end_time.strftime(plot_def.time_format)})"
    )

    all_lines = []
    for lines in emoji_lines.values():
        all_lines.extend(lines)
    all_labels = []
    for labels in emoji_labels.values():
        all_labels.extend(labels)

    ax.legend(all_lines, all_labels, loc="upper left")
    ax.set_xlabel("Time")
    ax.set_ylabel("Count")

    # Format x-axis
    ax.xaxis.set_major_formatter(dates.DateFormatter(plot_def.time_format))
    ax.xaxis.set_major_locator(plot_def.locator)
    ax.tick_params(axis="x", labelrotation=45)
    ax.grid(True, alpha=0.3)

    pyplot.tight_layout()

    return fig



"""Database model exceptions for logdetective."""


class KojiTaskNotFoundError(Exception):
    """Exception raised when a koji task is not found"""


class KojiTaskNotAnalyzedError(Exception):
    """Exception raised when a koji task analysis is still in progress"""


class KojiTaskAnalysisTimeoutError(Exception):
    """Exception raised when a koji task analysis has timed out"""



import asyncio
from collections import defaultdict
import datetime
from logging import BASIC_FORMAT
from typing import List, Dict, Optional
from pydantic import (
    BaseModel,
    Field,
    model_validator,
    field_validator,
    NonNegativeFloat,
    HttpUrl,
)

import aiohttp

from aiolimiter import AsyncLimiter
from gitlab import Gitlab
import koji

from logdetective.constants import (
    DEFAULT_TEMPERATURE,
    LLM_DEFAULT_MAX_QUEUE_SIZE,
    LLM_DEFAULT_REQUESTS_PER_MINUTE,
    SYSTEM_ROLE_DEFAULT,
    USER_ROLE_DEFAULT,
)

from logdetective.extractors import Extractor, DrainExtractor, CSGrepExtractor
from logdetective.utils import check_csgrep


class BuildLog(BaseModel):
    """Model of data submitted to API."""

    url: str


class JobHook(BaseModel):
    """Model of Job Hook events sent from GitLab.
    Full details of the specification are available at
    https://docs.gitlab.com/user/project/integrations/webhook_events/#job-events
    This model implements only the fields that we care about. The webhook
    sends many more fields that we will ignore."""

    # The unique job ID on this GitLab instance.
    build_id: int

    # The identifier of the job. We only care about 'build_rpm' and
    # 'build_centos_stream_rpm' jobs.
    build_name: str = Field(pattern=r"^build.*rpm$")

    # A string representing the job status. We only care about 'failed' jobs.
    build_status: str = Field(pattern=r"^failed$")

    # The kind of webhook message. We are only interested in 'build' messages
    # which represents job tasks in a pipeline.
    object_kind: str = Field(pattern=r"^build$")

    # The unique ID of the enclosing pipeline on this GitLab instance.
    pipeline_id: int

    # The unique ID of the project triggering this event
    project_id: int


class EmojiMergeRequest(BaseModel):
    """Model of the 'merge_request' subsection of Emoji webhook messages.
    This model implements only the fields that we care about. The webhook
    sends many more fields that we will ignore."""

    # The identifier of the target project
    target_project_id: int

    # The internal identifier (relative to the target project)
    iid: int


class EmojiHook(BaseModel):
    """Model of Job Hook events sent from GitLab.
    Full details of the specification are available at
    https://docs.gitlab.com/user/project/integrations/webhook_events/#job-events
    This model implements only the fields that we care about. The webhook
    sends many more fields that we will ignore."""

    # The kind of webhook message. We are only interested in 'emoji' messages
    # which represents awarding or revoking emoji reactions on notes.
    object_kind: str = Field(pattern=r"^emoji$")

    # Information about the merge request this emoji applies to, if any.
    merge_request: EmojiMergeRequest = Field(default=None)


class SnippetAnalysis(BaseModel):
    """Model of snippet analysis from LLM."""

    text: str = Field(description="Analysis of log snippet contents.")


class RatedSnippetAnalysis(SnippetAnalysis):
    """Model for rated snippet analysis. This model is used to generate
    json schema for inference with structured output."""

    relevance: int = Field(
        ge=0,
        le=100,
        description="Estimate of likelyhood that snippet contains an error, "
        "with 0 standing for completely unlikely, 100 for absolutely certain.",
    )


class Explanation(BaseModel):
    """Model of snippet or general log explanation from Log Detective"""

    text: str
    logprobs: Optional[List[Dict]] = None

    def __str__(self):
        """Return text of the Explanation"""
        return self.text


class AnalyzedSnippet(BaseModel):
    """Model for snippets already processed by Log Detective.

    explanation: LLM output in form of plain text and logprobs dictionary
    text: original snippet text
    line_number: location of snippet in original log
    """

    explanation: SnippetAnalysis | RatedSnippetAnalysis
    text: str
    line_number: int


class Response(BaseModel):
    """Model of data returned by Log Detective API

    explanation: Explanation
    response_certainty: float
    """

    explanation: Explanation
    response_certainty: float


class StagedResponse(Response):
    """Model of data returned by Log Detective API when called when staged response
    is requested. Contains list of reponses to prompts for individual snippets.

    explanation: Explanation
    response_certainty: float
    snippets: list of AnalyzedSnippet objects
    """

    snippets: List[AnalyzedSnippet]


class KojiStagedResponse(BaseModel):
    """Model of data returned by Log Detective API when called when a Koji build
    analysis is requested. Contains list of reponses to prompts for individual
    snippets.
    """

    task_id: int
    log_file_name: str
    response: StagedResponse


class InferenceConfig(BaseModel):  # pylint: disable=too-many-instance-attributes
    """Model for inference configuration of logdetective server."""

    max_tokens: int = -1
    log_probs: bool = True
    url: str = ""
    # OpenAI client library requires a string to be specified for API token
    # even if it is not checked on the server side
    api_token: str = "None"
    model: str = ""
    temperature: NonNegativeFloat = DEFAULT_TEMPERATURE
    max_queue_size: int = LLM_DEFAULT_MAX_QUEUE_SIZE
    http_timeout: float = 5.0
    user_role: str = USER_ROLE_DEFAULT
    system_role: str = SYSTEM_ROLE_DEFAULT
    llm_api_timeout: float = 15.0
    _http_session: aiohttp.ClientSession = None
    _limiter: AsyncLimiter = AsyncLimiter(LLM_DEFAULT_REQUESTS_PER_MINUTE)

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.max_tokens = data.get("max_tokens", -1)
        self.log_probs = data.get("log_probs", True)
        self.url = data.get("url", "")
        self.http_timeout = data.get("http_timeout", 5.0)
        self.api_token = data.get("api_token", "None")
        self.model = data.get("model", "default-model")
        self.temperature = data.get("temperature", DEFAULT_TEMPERATURE)
        self.max_queue_size = data.get("max_queue_size", LLM_DEFAULT_MAX_QUEUE_SIZE)
        self.user_role = data.get("user_role", USER_ROLE_DEFAULT)
        self.system_role = data.get("system_role", SYSTEM_ROLE_DEFAULT)
        self._requests_per_minute = data.get(
            "requests_per_minute", LLM_DEFAULT_REQUESTS_PER_MINUTE
        )
        self.llm_api_timeout = data.get("llm_api_timeout", 15.0)
        self._limiter = AsyncLimiter(self._requests_per_minute)

    def __del__(self):
        # Close connection when this object is destroyed
        if self._http_session:
            try:
                loop = asyncio.get_running_loop()
                loop.create_task(self._http_session.close())
            except RuntimeError:
                # No loop running, so create one to close the session
                loop = asyncio.new_event_loop()
                loop.run_until_complete(self._http_session.close())
                loop.close()
            except Exception:  # pylint: disable=broad-exception-caught
                # We should only get here if we're shutting down, so we don't
                # really care if the close() completes cleanly.
                pass

    def get_http_session(self):
        """Return the internal HTTP session so it can be used to contect the
        LLM server. May be used as a context manager."""

        # Create the session on the first attempt. We need to do this "lazily"
        # because it needs to happen once the event loop is running, even
        # though the initialization itself is synchronous.
        if not self._http_session:
            self._http_session = aiohttp.ClientSession(
                base_url=self.url,
                timeout=aiohttp.ClientTimeout(
                    total=self.http_timeout,
                    connect=3.07,
                ),
            )

        return self._http_session

    def get_limiter(self):
        """Return the limiter object so it can be used as a context manager"""
        return self._limiter


class ExtractorConfig(BaseModel):
    """Model for extractor configuration of logdetective server."""

    max_clusters: int = 8
    verbose: bool = False
    max_snippet_len: int = 2000
    csgrep: bool = False

    _extractors: List[Extractor] = []

    def _setup_extractors(self):
        """Initialize extractors with common settings."""
        self._extractors = [
            DrainExtractor(
                verbose=self.verbose,
                max_snippet_len=self.max_snippet_len,
                max_clusters=self.max_clusters,
            )
        ]

        if self.csgrep:
            self._extractors.append(
                CSGrepExtractor(
                    verbose=self.verbose,
                    max_snippet_len=self.max_snippet_len,
                )
            )

    def __init__(self, data: Optional[dict] = None):
        super().__init__(data=data)

        if data is None:
            self._setup_extractors()
            return

        self.max_clusters = data.get("max_clusters", 8)
        self.verbose = data.get("verbose", False)
        self.max_snippet_len = data.get("max_snippet_len", 2000)
        self.csgrep = data.get("csgrep", False)

        self._setup_extractors()

    def get_extractors(self) -> List[Extractor]:
        """Return list of initialized extractors, each will be applied in turn
        on original log text to retrieve snippets."""
        return self._extractors

    @field_validator("csgrep", mode="after")
    @classmethod
    def validate_csgrep(cls, value: bool) -> bool:
        """Verify that csgrep is available if requested."""
        if not check_csgrep():
            raise ValueError(
                "Requested csgrep extractor but `csgrep` binary is not in the PATH"
            )
        return value


class GitLabInstanceConfig(BaseModel):  # pylint: disable=too-many-instance-attributes
    """Model for GitLab configuration of logdetective server."""

    name: str = None
    url: str = None
    # Path to API of the gitlab instance, assuming `url` as prefix.
    api_path: str = None
    api_token: str = None

    # This is a list to support key rotation.
    # When the key is being changed, we will add the new key as a new entry in
    # the configuration and then remove the old key once all of the client
    # webhook configurations have been updated.
    # If this option is left empty or unspecified, all requests will be
    # considered authorized.
    webhook_secrets: Optional[List[str]] = None

    timeout: float = 5.0
    _conn: Gitlab = None
    _http_session: aiohttp.ClientSession = None

    # Maximum size of artifacts.zip in MiB. (default: 300 MiB)
    max_artifact_size: int = 300 * 1024 * 1024

    def __init__(self, name: str, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.name = name
        self.url = data.get("url", "https://gitlab.com")
        self.api_path = data.get("api_path", "/api/v4")
        self.api_token = data.get("api_token", None)
        self.webhook_secrets = data.get("webhook_secrets", None)
        self.max_artifact_size = int(data.get("max_artifact_size", 300)) * 1024 * 1024

        self.timeout = data.get("timeout", 5.0)
        self._conn = Gitlab(
            url=self.url,
            private_token=self.api_token,
            timeout=self.timeout,
        )

    def get_connection(self):
        """Get the Gitlab connection object"""
        return self._conn

    def get_http_session(self):
        """Return the internal HTTP session so it can be used to contect the
        Gitlab server. May be used as a context manager."""

        # Create the session on the first attempt. We need to do this "lazily"
        # because it needs to happen once the event loop is running, even
        # though the initialization itself is synchronous.
        if not self._http_session:
            self._http_session = aiohttp.ClientSession(
                base_url=self.url,
                headers={"Authorization": f"Bearer {self.api_token}"},
                timeout=aiohttp.ClientTimeout(
                    total=self.timeout,
                    connect=3.07,
                ),
            )

        return self._http_session

    def __del__(self):
        # Close connection when this object is destroyed
        if self._http_session:
            try:
                loop = asyncio.get_running_loop()
                loop.create_task(self._http_session.close())
            except RuntimeError:
                # No loop running, so create one to close the session
                loop = asyncio.new_event_loop()
                loop.run_until_complete(self._http_session.close())
                loop.close()
            except Exception:  # pylint: disable=broad-exception-caught
                # We should only get here if we're shutting down, so we don't
                # really care if the close() completes cleanly.
                pass


class GitLabConfig(BaseModel):
    """Model for GitLab configuration of logdetective server."""

    instances: Dict[str, GitLabInstanceConfig] = {}

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        for instance_name, instance_data in data.items():
            instance = GitLabInstanceConfig(instance_name, instance_data)
            self.instances[instance.url] = instance


class KojiInstanceConfig(BaseModel):
    """Model for Koji configuration of logdetective server."""

    name: str = ""
    xmlrpc_url: str = ""
    tokens: List[str] = []

    _conn: Optional[koji.ClientSession] = None
    _callbacks: defaultdict[int, set[str]] = defaultdict(set)

    def __init__(self, name: str, data: Optional[dict] = None):
        super().__init__()

        self.name = name
        if data is None:
            # Set some reasonable defaults
            self.xmlrpc_url = "https://koji.fedoraproject.org/kojihub"
            self.tokens = []
            self.max_artifact_size = 1024 * 1024
            return

        self.xmlrpc_url = data.get(
            "xmlrpc_url", "https://koji.fedoraproject.org/kojihub"
        )
        self.tokens = data.get("tokens", [])

    def get_connection(self):
        """Get the Koji connection object"""
        if not self._conn:
            self._conn = koji.ClientSession(self.xmlrpc_url)
        return self._conn

    def register_callback(self, task_id: int, callback: str):
        """Register a callback for a task"""
        self._callbacks[task_id].add(callback)

    def clear_callbacks(self, task_id: int):
        """Unregister a callback for a task"""
        try:
            del self._callbacks[task_id]
        except KeyError:
            pass

    def get_callbacks(self, task_id: int) -> set[str]:
        """Get the callbacks for a task"""
        return self._callbacks[task_id]


class KojiConfig(BaseModel):
    """Model for Koji configuration of logdetective server."""

    instances: Dict[str, KojiInstanceConfig] = {}
    analysis_timeout: int = 15
    max_artifact_size: int = 300 * 1024 * 1024

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        # Handle analysis_timeout with default 15
        self.analysis_timeout = data.get("analysis_timeout", 15)

        # Handle max_artifact_size with default 300
        self.max_artifact_size = data.get("max_artifact_size", 300) * 1024 * 1024

        # Handle instances dictionary
        instances_data = data.get("instances", {})
        for instance_name, instance_data in instances_data.items():
            self.instances[instance_name] = KojiInstanceConfig(
                instance_name, instance_data
            )


class LogConfig(BaseModel):
    """Logging configuration"""

    name: str = "logdetective"
    level_stream: str | int = "INFO"
    level_file: str | int = "INFO"
    path: str | None = None
    format: str = BASIC_FORMAT

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.name = data.get("name", "logdetective")
        self.level_stream = data.get("level_stream", "INFO").upper()
        self.level_file = data.get("level_file", "INFO").upper()
        self.path = data.get("path")
        self.format = data.get("format", BASIC_FORMAT)


class GeneralConfig(BaseModel):
    """General config options for Log Detective"""

    packages: List[str] = None
    excluded_packages: List[str] = None
    devmode: bool = False
    sentry_dsn: HttpUrl | None = None
    collect_emojis_interval: int = 60 * 60  # seconds
    top_k_snippets: int = 0

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.packages = data.get("packages", [])
        self.excluded_packages = data.get("excluded_packages", [])
        self.devmode = data.get("devmode", False)
        self.sentry_dsn = data.get("sentry_dsn")
        self.collect_emojis_interval = data.get(
            "collect_emojis_interval", 60 * 60
        )  # seconds
        self.top_k_snippets = data.get("top_k_snippets", 0)


class Config(BaseModel):
    """Model for configuration of logdetective server."""

    log: LogConfig = LogConfig()
    inference: InferenceConfig = InferenceConfig()
    snippet_inference: InferenceConfig = InferenceConfig()
    # TODO(jpodivin): Extend to work with multiple extractor configs
    extractor: ExtractorConfig = ExtractorConfig()
    gitlab: GitLabConfig = GitLabConfig()
    koji: KojiConfig = KojiConfig()
    general: GeneralConfig = GeneralConfig()

    def __init__(self, data: Optional[dict] = None):
        super().__init__()

        if data is None:
            return

        self.log = LogConfig(data.get("log"))
        self.inference = InferenceConfig(data.get("inference"))
        self.extractor = ExtractorConfig(data.get("extractor"))
        self.gitlab = GitLabConfig(data.get("gitlab"))
        self.koji = KojiConfig(data.get("koji"))
        self.general = GeneralConfig(data.get("general"))

        if snippet_inference := data.get("snippet_inference", None):
            self.snippet_inference = InferenceConfig(snippet_inference)
        else:
            self.snippet_inference = self.inference


class TimePeriod(BaseModel):
    """Specification for a period of time.

    If no indication is given
    it falls back to a 2 days period of time.

    Can't be smaller than a hour"""

    weeks: Optional[int] = None
    days: Optional[int] = None
    hours: Optional[int] = None

    @model_validator(mode="before")
    @classmethod
    def check_exclusive_fields(cls, data):
        """Check that only one key between weeks, days and hours is defined"""
        if isinstance(data, dict):
            how_many_fields = sum(
                1
                for field in ["weeks", "days", "hours"]
                if field in data and data[field] is not None
            )

            if how_many_fields == 0:
                data["days"] = 2  # by default fallback to a 2 days period

            if how_many_fields > 1:
                raise ValueError("Only one of months, weeks, days, or hours can be set")

        return data

    @field_validator("weeks", "days", "hours")
    @classmethod
    def check_positive(cls, v):
        """Check that the given value is positive"""
        if v is not None and v <= 0:
            raise ValueError("Time period must be positive")
        return v

    def get_time_period(self) -> datetime.timedelta:
        """Get the period of time represented by this input model.

        Returns:
            datetime.timedelta: The time period as a timedelta object.
        """
        delta = None
        if self.weeks:
            delta = datetime.timedelta(weeks=self.weeks)
        elif self.days:
            delta = datetime.timedelta(days=self.days)
        elif self.hours:
            delta = datetime.timedelta(hours=self.hours)
        return delta

    def get_period_start_time(
        self, end_time: datetime.datetime = None
    ) -> datetime.datetime:
        """Calculate the start time of this period based on the end time.

        Args:
            end_time (datetime.datetime, optional): The end time of the period.
                Defaults to current UTC time if not provided.

        Returns:
            datetime.datetime: The start time of the period.
        """
        time = end_time or datetime.datetime.now(datetime.timezone.utc)
        if time.tzinfo is None:
            end_time = end_time.replace(tzinfo=datetime.timezone.utc)
        return time - self.get_time_period()



from typing import List
from importlib.metadata import version

import aiohttp
from fastapi import HTTPException
from fastapi.responses import Response as BasicResponse

from logdetective.constants import SNIPPET_DELIMITER
from logdetective.server.config import LOG
from logdetective.server.exceptions import LogDetectiveConnectionError
from logdetective.server.models import AnalyzedSnippet, RatedSnippetAnalysis


def format_analyzed_snippets(snippets: list[AnalyzedSnippet]) -> str:
    """Format snippets for submission into staged prompt."""
    summary = f"\n{SNIPPET_DELIMITER}\n".join(
        [f"[{e.text}] at line [{e.line_number}]: [{e.explanation}]" for e in snippets]
    )
    return summary


def connection_error_giveup(details: dict) -> None:
    """Too many connection errors, give up.
    """
    LOG.error("Too many connection errors, giving up. %s", details["exception"])
    raise LogDetectiveConnectionError() from details["exception"]


def should_we_giveup(exc: aiohttp.ClientResponseError) -> bool:
    """From backoff's docs:

    > a function which accepts the exception and returns
    > a truthy value if the exception should not be retried
    """
    LOG.info("Should we give up on retrying error %s", exc)
    return exc.status < 400


def we_give_up(details: dict):
    """Retries didn't work (or we got a different exc)
    we give up and raise proper 500 for our API endpoint
    """
    LOG.error("Last exception: %s", details["exception"])
    LOG.error("Inference error: %s", details["args"])
    raise HTTPException(500, "Request to the inference API failed")


def select_relevance(snippet: AnalyzedSnippet) -> float:
    """Retrieve relevance value from structure, if there is one."""
    if not isinstance(snippet.explanation, RatedSnippetAnalysis):
        LOG.exception("Only rated snippets can be ordered by relevance.")
        raise ValueError
    return snippet.explanation.relevance


def select_line_number(explanation: AnalyzedSnippet) -> int:
    """Returns line number of original snippet."""
    return explanation.line_number


def filter_snippets(
    processed_snippets: List[AnalyzedSnippet], top_k: int
) -> List[AnalyzedSnippet]:
    """Filter snippets according to criteria in config while keeping them ordered by line number.
    If all snippets recieved the same score, return them all.
    AnalyzedSnippet objects must have `explanation` attribute set to `RatedSnippetAnalysis`,
    otherwise raise `ValueError`."""

    if top_k >= len(processed_snippets):
        LOG.warning(
            "The `top-k` parameter >= number of original snippets, skipping filtering."
        )
        return processed_snippets

    # Sorting invokes `select_relevance` which also tests if objects actually
    # have the score assigned. Otherwise it raises exception.
    processed_snippets = sorted(processed_snippets, key=select_relevance, reverse=True)

    # Check for failure mode when all snippets have
    # the same relevance. In such cases there is no point in filtering
    # and all snippets are returned.
    max_relevance = processed_snippets[0].explanation.relevance
    min_relevance = processed_snippets[-1].explanation.relevance

    LOG.info(
        "Analyzed snippets sorted. Max relevance: %d Min relevance: %e",
        max_relevance,
        min_relevance,
    )
    if max_relevance == min_relevance:
        LOG.warning("All snippets recieved the same rating. Filtering disabled.")
        return processed_snippets

    processed_snippets = processed_snippets[:top_k]

    # Re-sorting snippets by line number
    processed_snippets = sorted(processed_snippets, key=select_line_number)

    return processed_snippets


def construct_final_prompt(formatted_snippets: str, prompt_template: str) -> str:
    """Create final prompt from processed snippets and csgrep output, if it is available."""

    final_prompt = prompt_template.format(formatted_snippets)
    return final_prompt


def get_version() -> BasicResponse:
    """Obtain the version number using importlib"""
    return BasicResponse(content=version('logdetective'))
